"""
Created on Fri Jan  4 13:19:28 2019

Based on code from Fluid Dynamics sixth edition section 6.3 (code 3)

Solution for Unsteady Two-Dimensional Navier-Stokes equations in vorticity-stream function form, using an explicit forward in time, centered in space scheme

Wind tunnel simulation expanded by repeated boundary conditions on horizontal boundary and net zero flux on vertical boundaries

Succesive Over Relaxation (SOR) is used for generation of stream function, assuming uniform resolution

@author: fbullard
"""

from scipy import zeros, linspace, meshgrid, sqrt
from numpy import sum as sumpy
import matplotlib.pyplot as plt

    #parameters for SOR
scale=1 ; Ny=60 ; Nx=(Ny*scale)-scale+1 ; MaxStep=1000 ; Visc=.5 ; dt=0.00005 ; t = 0.0; MaxIt=600 ; Beta=1.5 ; MaxErr=0.001 ;

    #functions and intitial conditions
sf=zeros([Ny, Nx]) ; vt=zeros([Ny, Nx]) ; vto=zeros([Ny,Nx]) ; 
Ui = .1; height=2.0

    #for contour mapping. h is also used for boundary conditions on stream function
x=zeros([Ny,Nx]) ; y=zeros([Ny,Nx]) ; h=height/(Ny-1);
    
    # for creation of shapes
shape = zeros([ Ny , Nx ])
xi = linspace(-(height*scale)/2,(height*scale)/2, Nx , endpoint = True)
yj = linspace(-height/2, height/2, Ny, endpoint = True)
xx, yy = meshgrid(xi, yj, indexing = 'xy')

for a in range(Nx):
    for b in range(Ny):
        if sqrt(xx[b,a]**2 + yy[b,a]**2) > 0.3:
            shape[b,a] = 1

    #could be replaced by meshgrid
for i in range(Nx):
    for j in range(Ny):
        x[j,i] = h*i
        y[j,i] = h*j
        
    #generate boundaries for stream function
for j in range(Ny):
    sf[j,0] = Ui*j*h
    sf[j,Nx-1] = Ui*j*h
    
for tstep in range(MaxStep):
    for SOR in range(MaxIt):
        vto = sf.copy()
        for i in range(1,Nx-1):
            for j in range(1,Ny-1):
                sf[j,i] = 0.25*Beta*(sf[j,i+1] + sf[j,i-1] + sf[j+1,i] + sf[j-1,i] + h*h*vt[j,i]) + (1.0-Beta)*sf[j,i]
        
            #for repeated boundary condition solve bottom with solution of third from top then set bottom two rows to top two rows
            
        for i in range(1,Nx-1):
            sf[Ny-1, i] = 0.25*Beta*(sf[Ny-1,i+1] + sf[Ny-1,i-1] + sf[2,i] + sf[Ny-2,i] + h*h*vt[j,i]) + (1.0-Beta)*sf[j,i]
        sf[:2,1:Nx-1] = sf[Ny-2:Ny,1:Nx-1]
        sf *= shape
    
            #check for convergence of the stream function given some maximum error and steps
            
        error = sumpy(abs(sf - vto))
        if error <= MaxErr:
            break
            
        #generate vorticity function for boundaries
        
    vt[1:Ny-1,0] = (2.0*(sf[1:Ny-1,0] - sf[1:Ny-1,1]))/(h*h)
    vt[1:Ny-1,Nx-1] = (2.0*(sf[1:Ny-1,Nx-1] - sf[1:Ny-1,Nx-2]))/(h*h)
    vt[0,1:Nx-1] = (2.0*(sf[0,1:Nx-1] - sf[1,1:Nx-1]))/(h*h)
    vt[Ny-1,1:Nx-1] = (2.0*(sf[0,1:Nx-1] - sf[1,1:Nx-1]))/(h*h)
    vto = vt.copy()
    
    #print(tstep)
    
    for i in range(1,Nx-1):
        for j in range(1,Ny-1):
            vt[j,i] = vt[j,i]+dt*(-0.25*((sf[j+1,i]-sf[j-1,i])*(vto[j,i+1]-vto[j,i-1]) - (sf[j,i+1]-sf[j,i-1])*(vto[j+1,i]-vto[j-1,i]))/(h*h) \
            +Visc*(vto[j,i+1]+vto[j,i-1]+vto[j+1,i]+vto[j-1,i]-4.0*vto[j,i])/(h*h) )
    vt *= shape
    
    t = t + dt
plt.subplot(121)
plt.contour(x,y,vt,40)
plt.subplot(122)
plt.contour(x,y,sf)
plt.show()